using System;
using DerivativeFunctions;
using LinearAlgebra;
using Steppers;
using NumericsOfDifferentialEquations;
using Graphing;

namespace Experiments
{
	/// <summary>
	/// A complete settings for an experiment with differential equations.
	/// </summary>
	public interface Experiment
	{
		string Step();
		void Dispose();
	}

	public class CompareMethods : Experiment
	{
		private bool showXY;
		private GraphWindow[] w;
		private Stepper[] s;
		private DerivativeFunction f;
		private VectorN[] v;
		private double t;
		private double h;

		public CompareMethods(bool showXY, DerivativeFunction f, double h, VectorN vStart, params Stepper[] s)
		{
			this.showXY = showXY;
			this.f = f;
			t = 0.0;
			this.h = h;
			this.s = s;
			w = GraphWindow.Create(s.Length);
			
			v = new VectorN[s.Length];
			for(int i = 0; i < s.Length; i++)
			{
				v[i] = vStart;
			}
		}

		public void Dispose()
		{
			foreach(GraphWindow win in w)
			{
				win.Dispose();
			}
		}

		public string Step()
		{
			for(int i = 0; i < s.Length; i++)
			{
				v[i] = s[i].GetNext(f, t, v[i], h);
				if(showXY)
				{
					w[i].Append(v[i].Get(0), v[i].Get(2));
				}
				else
				{
					w[i].Append(t + h, v[i].Get(0));
				}
			}
			double oldT = t;
			t += h;
			return "t = " + oldT.ToString("f");
		}
	}

	public class ShowStabilityRegion : Experiment
	{
		private GraphWindow[] w;
		private Stepper[] s;
		private int number;
		private Random rand;

		public ShowStabilityRegion(params Stepper[] s)
		{
			this.s = s;
			w = GraphWindow.Create(s.Length, 2);
			rand = new Random();
		}

		public void Dispose()
		{
			foreach(GraphWindow win in w)
			{
				win.Dispose();
			}
		}

		public string Step()
		{
			VectorN v = new VectorN(1.0, 0.0);
			double real = -3.0*rand.NextDouble();
			double imag = 6.0*rand.NextDouble() - 3.0;
			double[,] m = new double[,] {{real, -imag}, {imag, real}};
			MatrixN mat = new MatrixN(ref m);
			DerivativeFunction f = new MatrixMultiplication(ref mat);
			for(int i = 0; i < s.Length; i++)
			{
				double length = s[i].GetNext(f, 0.0, v, 1.0).Length();
				if(length < 1.0)
				{
					w[i].Append(real, imag);
				}
			}
			number++;
			return "N = " + number.ToString("d");
		}
	}

	public class TestPrecision : Experiment
	{
		private GraphWindow[] w;
		private DerivativeFunction f;
		private double tEnd;
		private double h;
		private VectorN vStart;
		private double endValue;
		private Stepper[] s;

		public TestPrecision(DerivativeFunction f, double tEnd, double hStart, VectorN vStart, double endValue, params Stepper[] s)
		{
			this.f = f;
			this.tEnd = tEnd;
			this.h = hStart;
			this.vStart = vStart;
			this.endValue = endValue;
			this.s = s;
			w = GraphWindow.Create(s.Length);
		}

		public void Dispose()
		{
			foreach(GraphWindow win in w)
			{
				win.Dispose();
			}
		}

		public string Step()
		{
			for(int i = 0; i < s.Length; i++)
			{
				double t = 0.0;
				VectorN v = vStart;
				while(t < tEnd)
				{
					v = s[i].GetNext(f, t, v, h);
					t += h;
				}
				w[i].InsertFront(Math.Log10(h), Math.Log10(Math.Abs(v.Get(0)-endValue)));
			}

			double oldH = h;
			h *= 0.5;
			return "h = " + oldH.ToString("e2");
		}
	}

	public class CompareDistanceGrowth : Experiment
	{
		private GraphWindow[] w;
		private DerivativeFunction f;
		private double t;
		private double h;
		private VectorN v1, v2;
		private Stepper s;

		public CompareDistanceGrowth(DerivativeFunction f, double h, VectorN vStart1, VectorN vStart2, Stepper s)
		{
			this.f = f;
			t = 0.0;
			this.h = h;
			this.v1 = vStart1;
			this.v2 = vStart2;
			this.s = s;
			w = GraphWindow.Create(4);
		}

		public void Dispose()
		{
			foreach(GraphWindow win in w)
			{
				win.Dispose();
			}
		}

		public string Step()
		{
			v1 = s.GetNext(f, t, v1, h);
			v2 = s.GetNext(f, t, v2, h);
			w[0].Append(v1.Get(0), v1.Get(1));
			w[2].Append(v2.Get(0), v2.Get(1));
			double distance = (v1-v2).Length();
			w[1].Append(t, distance);
			w[3].Append(t, Math.Log10(distance));

			double oldT = t;
			t += h;
			return "t = " + oldT.ToString("f");
		}
	}
}
