using System;

namespace LinearAlgebra
{
	/// <summary>
	/// A square matrix of arbitray size.
	/// </summary>
	public struct MatrixN
	{
		private double[,] m;

		public MatrixN(int size)
		{
			m = new double[size, size];
		}

		public MatrixN(ref double[,] elems)
		{
			if(elems.GetLength(0) != elems.GetLength(1))
			{
				throw new Exception("Non-square matrix.");
			}
			m = elems;
		}

		public int Size
		{
			get
			{
				return m.GetLength(0);
			}
		}

		public double Get(int i, int j)
		{
			return m[i, j];
		}

		public void Set(int i, int j, double value)
		{
			m[i, j] = value;
		}

		private static void checkDimensions(ref MatrixN m1, ref MatrixN m2)
		{
			if(m1.Size != m2.Size)
			{
				throw new Exception("Dimension mismatch.");
			}
		}

		public static MatrixN operator + (MatrixN m1, MatrixN m2)
		{
			checkDimensions(ref m1, ref m2);
			double[,] m = new double[m1.Size, m1.Size];
			for(int i = 0; i < m1.Size; i++)
			{
				for(int j = 0; j < m1.Size; j++)
				{
					m[i, j] = m1.m[i, j] + m2.m[i, j];
				}
			}
			return new MatrixN(ref m);
		}

		public static MatrixN operator - (MatrixN m1, MatrixN m2)
		{
			checkDimensions(ref m1, ref m2);
			double[,] m = new double[m1.Size, m1.Size];
			for(int i = 0; i < m1.Size; i++)
			{
				for(int j = 0; j < m1.Size; j++)
				{
					m[i, j] = m1.m[i, j] - m2.m[i, j];
				}
			}
			return new MatrixN(ref m);
		}

		public static MatrixN operator * (double f, MatrixN m1)
		{
			double[,] n = (double[,])m1.m.Clone();
			for(int i = 0; i < m1.Size; i++)
			{
				for(int j = 0; j < m1.Size; j++)
				{
					n[i, j] *= f;
				}
			}
			return new MatrixN(ref n);
		}

		public static MatrixN operator * (MatrixN m1, double f)
		{
			return f*m1;
		}

		public static MatrixN operator / (MatrixN m1, double f)
		{
			double[,] n = (double[,])m1.m.Clone();
			for(int i = 0; i < m1.Size; i++)
			{
				for(int j = 0; j < m1.Size; j++)
				{
					n[i, j] /= f;
				}
			}
			return new MatrixN(ref n);
		}

		public MatrixN Invert()
		{
			if(Size == 1)
			{
				double[,] m = new double[,] {{1.0/Get(0, 0)}};
				return new MatrixN(ref m);
			}
			else if(Size == 2)
			{
				double[,] m = new double[,] {{Get(1, 1), -Get(1, 0)}, {Get(0, 1), Get(0, 0)}};
				return new MatrixN(ref m) / (Get(0, 0)*Get(1, 1) - Get(0, 1)*Get(0, 0));
			}
			else
			{
				throw new Exception("Only implemented for dimensions 1 and 2.");
			}
		}

		public static VectorN operator * (MatrixN m1, VectorN v1)
		{
			if(m1.Size != v1.Size)
			{
				throw new Exception("Dimension mismatch.");
			}

			double[] a = new double[v1.Size];
			for(int i = 0; i < m1.Size; i++)
			{
				for(int j = 0; j < m1.Size; j++)
				{
					a[i] += m1.Get(i, j) * v1.Get(j);
				}
			}

			return new VectorN(ref a);
		}

		public static MatrixN Identity(int size)
		{
			MatrixN m1 = new MatrixN(size);
			for(int i = 0; i < size; i++)
			{
				m1.Set(i, i, 1.0);
			}
			return m1;
		}
	}
}
